@@ -1,10 +1,7 @@
 #!perl -w
 use strict;
-
-use Text::Xslate;
-BEGIN {
-    eval{ require HTML::FillInForm::Lite::Compat };
-}
+use Text::Xslate qw(mark_raw);
+BEGIN { eval { require HTML::FillInForm::Lite::Compat } }
 use HTML::FillInForm;
 
 sub fillinform {
@@ -12,20 +9,22 @@ sub fillinform {
 
     return sub {
         my($html) = @_;
-        return HTML::FillInForm->fill(\$html, $q);
+        return mark_raw(HTML::FillInForm->fill(\$html, $q));
     };
 }
 
 my $tx  = Text::Xslate->new(
-    cache => 0,
     function => {
         fillinform => \&fillinform,
     },
 );
 
-print $tx->render_string(<<'T', { q => { foo => "<filled value>" } });
-FillInForm
-: block form | fillinform($q) | raw -> {
+my %vars = (
+    q => { foo => "<filled value>" },
+);
+print $tx->render_string(<<'T', \%vars);
+FillInForm:
+: block form | fillinform($q) -> {
 <form>
 <input type="text" name="foo" />
 </form>
