@@ -4,7 +4,7 @@ use 5.008_001;
 use strict;
 use warnings;
 
-our $VERSION = '0.1039';
+our $VERSION = '0.1040';
 
 use Carp       ();
 use File::Spec ();
@@ -43,6 +43,7 @@ use Text::Xslate::Util qw(
     $NUMBER $STRING $DEBUG
     literal_to_value
     import_from
+    make_error
 );
 
 BEGIN {
@@ -180,12 +181,12 @@ sub flush_memory_cache {
     return;
 }
 
-sub load_string { # for <input>
+sub load_string { # for <string>
     my($self, $string) = @_;
     if(not defined $string) {
         $self->_error("LoadError: Template string is not given");
     }
-    $self->{string} = $string;
+    $self->{string_buffer} = $string;
     my $asm = $self->compile($string);
     $self->_assemble($asm, undef, undef, undef, undef);
     return $asm;
@@ -221,13 +222,12 @@ sub find_file {
     }
 
     if(not defined $orig_mtime) {
-        $self->_error("LoadError: Cannot find $file (path: @{$self->{path}})");
+        $self->_error("LoadError: Cannot find '$file' (path: @{$self->{path}})");
     }
 
     print STDOUT "  find_file: $fullpath (", ($cache_mtime || 0), ")\n" if _DUMP_LOAD_FILE;
 
     return {
-        file        => $file,
         fullpath    => $fullpath,
         cachepath   => $cachepath,
 
@@ -241,8 +241,8 @@ sub load_file {
 
     print STDOUT "load_file($file)\n" if _DUMP_LOAD_FILE;
 
-    if($file eq '<input>') { # simply reload it
-        return $self->load_string($self->{string});
+    if($file eq '<string>') { # simply reload it
+        return $self->load_string($self->{string_buffer});
     }
 
     my $fi = $self->find_file($file, $mtime);
@@ -285,8 +285,7 @@ sub _load_source {
     my $source = $self->slurp($fullpath);
 
     my $asm = $self->compile($source,
-        file     => $fi->{file},
-        fullpath => $fullpath,
+        file => $fullpath,
     );
 
     if($self->{cache} >= 1) {
@@ -428,7 +427,6 @@ sub _compiler {
     my $compiler = $self->{compiler};
 
     if(!ref $compiler){
-        $compiler ||= $self->compiler_class;
         require Any::Moose;
         Any::Moose::load_class($compiler);
 
@@ -454,9 +452,7 @@ sub compile {
 }
 
 sub _error {
-    shift;
-    unshift @_, 'Xslate: ';
-    goto &Carp::croak;
+    die make_error(@_);
 }
 
 sub dump :method {
@@ -473,7 +469,7 @@ Text::Xslate - High performance template engine
 
 =head1 VERSION
 
-This document describes Text::Xslate version 0.1039.
+This document describes Text::Xslate version 0.1040.
 
 =head1 SYNOPSIS
 
