@@ -11,7 +11,10 @@ our @EXPORT_OK = qw(
     html_escape escaped_string
     literal_to_value value_to_literal
     import_from
+    neat
     is_int any_in
+    read_around
+    make_error
     p
     $STRING $NUMBER $DEBUG
 );
@@ -35,6 +38,11 @@ our $NUMBER  = qr/ [+-]? (?:
 our $DEBUG;
 defined($DEBUG) or $DEBUG = $ENV{XSLATE} || '';
 
+our $DisplayWidth = 76;
+if($DEBUG =~ /display_width=(\d+)/) {
+    $DisplayWidth = $1;
+}
+
 require Text::Xslate; # load XS stuff
 
 sub mark_raw;    # XS
@@ -50,6 +58,21 @@ sub Text::Xslate::EscapedString::new {
     return Text::Xslate::Type::Raw->new($str);
 }
 
+sub neat {
+    my($s) = @_;
+    if ( defined $s ) {
+        if ( ref($s) || Scalar::Util::looks_like_number($s) ) {
+            return $s;
+        }
+        else {
+            return "'$s'";
+        }
+    }
+    else {
+        return 'nil';
+    }
+}
+
 sub is_int {
     my($s) = @_;
     my $i = do {
@@ -171,6 +194,65 @@ END_IMPORT
     return @funcs;
 }
 
+sub make_error {
+    my($self, $message, $file, $line, @extra) = @_;
+    if(ref $message) { # re-thrown form virtual machines
+        return ${$message};
+    }
+
+    my $lines = read_around($file, $line);
+    if($lines) {
+        $lines .= "\n" if $lines !~ /\n\z/xms;
+        $lines = '-' x $DisplayWidth . "\n"
+               . $lines
+               . '-' x $DisplayWidth . "\n";
+    }
+
+    local $Carp::CarpLevel = $Carp::CarpLevel + 1;
+    my $class = ref($self) || $self;
+    $message =~ s/\A \Q$class: \E//xms and $message .= "\t...";
+
+    if(defined $file) {
+        if(defined $line) {
+            unshift @extra, $line;
+        }
+        unshift @extra, ref($file) ? '<string>' : $file;
+    }
+
+    if(@extra) {
+        $message = Carp::shortmess(sprintf '%s (%s)',
+            $message, join(':', @extra));
+    }
+    else {
+        $message = Carp::shortmess($message);
+    }
+    return sprintf "%s: %s%s",
+        $class, $message, $lines;
+}
+
+sub read_around { # for error messages
+    my($file, $line, $around) = @_;
+
+    defined($file) && defined($line)       or return '';
+
+    $around = 1 if not defined $around;
+
+    open my $in, '<', $file or return '';
+    local $/ = "\n";
+    local $. = 0;
+
+    my $s = '';
+    while(defined(my $l = <$in>)) {
+        if($. >= ($line - $around)) {
+            $s .= $l;
+        }
+        if($. >= ($line + $around)) {
+            last;
+        }
+    }
+    return $s;
+}
+
 sub p { # for debugging
     require 'Data/Dumper.pm'; # we don't want to create its namespace
     my $dd = Data::Dumper->new([@_ == 1 ? @_ : \@_], ['*data']);
